/*
 * hp9000.c -- lightweight process initialisation for hp9000.
 * Copyright (C) 1991-6 Stephen Crane.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * author: Stephen Crane, (jsc@doc.ic.ac.uk), Department of Computing,
 * Imperial College of Science, Technology and Medicine, 180 Queen's
 * Gate, London SW7 2BZ, England.
 */

#include "lwp.h"

/*
 * initp -- initialise a new process's context.
 */
void initp (volatile struct pcb *volatile newp, void *sp)
{
	static registers *cpp;
	extern struct pcb *currp;

	if (!savep (currp->context)) {
		cpp = (jmp_buf *)&newp->context;
		asm volatile ("ldw	%0, %%sp": : "o" (sp));
		if (!savep (*cpp))
			restorep (currp->context);
		wrapp ();		
	}
}

int savep (registers jb) {
	/* save stack pointer and return program counter */
	asm ("stw	%sp, 4(%arg0)");
	asm ("stw	%rp, 8(%arg0)");

	/* save "callee save" registers */
	asm ("stw	%r3, 12(%arg0)");
	asm ("stw	%r4, 16(%arg0)");
	asm ("stw	%r5, 20(%arg0)");
	asm ("stw	%r6, 24(%arg0)");
	asm ("stw	%r7, 28(%arg0)");
	asm ("stw	%r8, 32(%arg0)");
	asm ("stw	%r9, 36(%arg0)");
	asm ("stw	%r10, 40(%arg0)");
	asm ("stw	%r11, 44(%arg0)");
	asm ("stw	%r12, 48(%arg0)");
	asm ("stw	%r13, 52(%arg0)");
	asm ("stw	%r14, 56(%arg0)");
	asm ("stw	%r15, 60(%arg0)");
	asm ("stw	%r16, 64(%arg0)");
	asm ("stw	%r17, 68(%arg0)");
	asm ("stw	%r18, 72(%arg0)");

	/* save "callee save" space register */
	asm volatile ("mfsp	%sr3, %r1");
	asm ("stw	%r1, 0(%arg0)");

	/* indicate "true return" from saved() */
	asm ("ldi	0, %ret0");

	asm (".LABEL _comefrom_");
}

void restorep (registers jb) {
	/* restore stack pointer and program counter */
	asm volatile ("ldw	4(%arg0), %sp");
	asm volatile ("ldw	8(%arg0), %rp");

	/* restore "callee save" space register */
	asm volatile ("ldw	0(%arg0), %r1");
	asm volatile ("mtsp	%r1, %sr3");

	/* restore "callee save" registers */
	asm volatile ("ldw	12(%arg0), %r3");
	asm volatile ("ldw	16(%arg0), %r4");
	asm volatile ("ldw	20(%arg0), %r5");
	asm volatile ("ldw	24(%arg0), %r6");
	asm volatile ("ldw	28(%arg0), %r7");
	asm volatile ("ldw	32(%arg0), %r8");
	asm volatile ("ldw	36(%arg0), %r9");
	asm volatile ("ldw	40(%arg0), %r10");
	asm volatile ("ldw	44(%arg0), %r11");
	asm volatile ("ldw	48(%arg0), %r12");
	asm volatile ("ldw	52(%arg0), %r13");
	asm volatile ("ldw	56(%arg0), %r14");
	asm volatile ("ldw	60(%arg0), %r15");
	asm volatile ("ldw	64(%arg0), %r16");
	asm volatile ("ldw	68(%arg0), %r17");
	asm volatile ("ldw	72(%arg0), %r18");

	/* warp to saved() to unwind the frame correctly */
	asm volatile ("bl	_comefrom_, %r0");	
	asm volatile ("ldi	1, %ret0");
}
